#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <assert.h>
#include <getopt.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>
#include <fcntl.h>
#include <vector>
#include <string>
#include <map>
#include <set>

#include "acceptor.h"
#include "config.h"
#include "markpool.h"
#include "metacube.h"
#include "parse.h"
#include "server.h"
#include "serverpool.h"
#include "input.h"
#include "stats.h"
#include "util.h"
#include "version.h"
#include "state.pb.h"

using namespace std;

ServerPool *servers = NULL;
volatile bool hupped = false;

void hup(int ignored)
{
	hupped = true;
}

CubemapStateProto collect_state(const timeval &serialize_start,
                                const vector<Acceptor *> acceptors,
                                const vector<Input *> inputs,
                                ServerPool *servers)
{
	CubemapStateProto state = servers->serialize();  // Fills streams() and clients().
	state.set_serialize_start_sec(serialize_start.tv_sec);
	state.set_serialize_start_usec(serialize_start.tv_usec);
	
	for (size_t i = 0; i < acceptors.size(); ++i) {
		state.add_acceptors()->MergeFrom(acceptors[i]->serialize());
	}

	for (size_t i = 0; i < inputs.size(); ++i) {
		state.add_inputs()->MergeFrom(inputs[i]->serialize());
	}

	return state;
}

// Find all port statements in the configuration file, and create acceptors for htem.
vector<Acceptor *> create_acceptors(
	const Config &config,
	map<int, Acceptor *> *deserialized_acceptors)
{
	vector<Acceptor *> acceptors;
	for (unsigned i = 0; i < config.acceptors.size(); ++i) {
		const AcceptorConfig &acceptor_config = config.acceptors[i];
		Acceptor *acceptor = NULL;
		map<int, Acceptor *>::iterator deserialized_acceptor_it =
			deserialized_acceptors->find(acceptor_config.port);
		if (deserialized_acceptor_it != deserialized_acceptors->end()) {
			acceptor = deserialized_acceptor_it->second;
			deserialized_acceptors->erase(deserialized_acceptor_it);
		} else {
			int server_sock = create_server_socket(acceptor_config.port, TCP_SOCKET);
			acceptor = new Acceptor(server_sock, acceptor_config.port);
		}
		acceptor->run();
		acceptors.push_back(acceptor);
	}

	// Close all acceptors that are no longer in the configuration file.
	for (map<int, Acceptor *>::iterator acceptor_it = deserialized_acceptors->begin();
	     acceptor_it != deserialized_acceptors->end();
	     ++acceptor_it) {
		acceptor_it->second->close_socket();
		delete acceptor_it->second;
	}

	return acceptors;
}

// Find all streams in the configuration file, and create inputs for them.
vector<Input *> create_inputs(const Config &config,
                              map<string, Input *> *deserialized_inputs)
{
	vector<Input *> inputs;
	for (unsigned i = 0; i < config.streams.size(); ++i) {
		const StreamConfig &stream_config = config.streams[i];
		if (stream_config.src.empty()) {
			continue;
		}

		string stream_id = stream_config.stream_id;
		string src = stream_config.src;

		Input *input = NULL;
		map<string, Input *>::iterator deserialized_input_it =
			deserialized_inputs->find(stream_id);
		if (deserialized_input_it != deserialized_inputs->end()) {
			input = deserialized_input_it->second;
			if (input->get_url() != src) {
				fprintf(stderr, "INFO: Stream '%s' has changed URL from '%s' to '%s', restarting input.\n",
					stream_id.c_str(), input->get_url().c_str(), src.c_str());
				input->close_socket();
				delete input;
				input = NULL;
			}
			deserialized_inputs->erase(deserialized_input_it);
		}
		if (input == NULL) {
			input = create_input(stream_id, src);
			if (input == NULL) {
				fprintf(stderr, "ERROR: did not understand URL '%s', clients will not get any data.\n",
				        src.c_str());
				continue;
			}
		}
		input->run();
		inputs.push_back(input);
	}
	return inputs;
}

void create_streams(const Config &config,
                    const set<string> &deserialized_stream_ids,
		    map<string, Input *> *deserialized_inputs)
{
	vector<MarkPool *> mark_pools;  // FIXME: leak
	for (unsigned i = 0; i < config.mark_pools.size(); ++i) {
		const MarkPoolConfig &mp_config = config.mark_pools[i];
		mark_pools.push_back(new MarkPool(mp_config.from, mp_config.to));
	}

	set<string> expecting_stream_ids = deserialized_stream_ids;
	for (unsigned i = 0; i < config.streams.size(); ++i) {
		const StreamConfig &stream_config = config.streams[i];
		if (deserialized_stream_ids.count(stream_config.stream_id) == 0) {
			servers->add_stream(stream_config.stream_id);
		}
		expecting_stream_ids.erase(stream_config.stream_id);

		if (stream_config.mark_pool != -1) {
			servers->set_mark_pool(stream_config.stream_id,
			                       mark_pools[stream_config.mark_pool]);
		}
	}

	// Warn about any servers we've lost.
	// TODO: Make an option (delete=yes?) to actually shut down streams.
	for (set<string>::const_iterator stream_it = expecting_stream_ids.begin();
	     stream_it != expecting_stream_ids.end();
	     ++stream_it) {
		string stream_id = *stream_it;
		fprintf(stderr, "WARNING: stream '%s' disappeared from the configuration file.\n",
			stream_id.c_str());
		fprintf(stderr, "         It will not be deleted, but clients will not get any new inputs.\n");
		if (deserialized_inputs->count(stream_id) != 0) {
			delete (*deserialized_inputs)[stream_id];
			deserialized_inputs->erase(stream_id);
		}
	}
}
	
bool dry_run_config(const std::string &argv0, const std::string &config_filename)
{
	char *argv0_copy = strdup(argv0.c_str());
	char *config_filename_copy = strdup(config_filename.c_str());

	pid_t pid = fork();
	switch (pid) {
	case -1:
		perror("fork()");
		free(argv0_copy);
		free(config_filename_copy);
		return false;
	case 0:
		// Child.
		execlp(argv0_copy, argv0_copy, "--test-config", config_filename_copy, NULL);
		perror(argv0_copy);
		_exit(1);
	default:
		// Parent.
		break;
	}
		
	free(argv0_copy);
	free(config_filename_copy);

	int status;
	pid_t err;
	do {
		err = waitpid(pid, &status, 0);
	} while (err == -1 && errno == EINTR);

	if (err == -1) {
		perror("waitpid()");
		return false;
	}	

	return (WIFEXITED(status) && WEXITSTATUS(status) == 0);
}

int main(int argc, char **argv)
{
	// Parse options.
	int state_fd = -1;
	bool test_config = false;
	for ( ;; ) {
		static const option long_options[] = {
			{ "state", required_argument, 0, 's' },
			{ "test-config", no_argument, 0, 't' },
		};
		int option_index = 0;
		int c = getopt_long (argc, argv, "s:t", long_options, &option_index);
     
		if (c == -1) {
			break;
		}
		switch (c) {
		case 's':
			state_fd = atoi(optarg);
			break;
		case 't':
			test_config = true;
			break;
		default:
			assert(false);
		}
	}

	string config_filename = "cubemap.config";
	if (optind < argc) {
		config_filename = argv[optind++];
	}

	Config config;
	if (!parse_config(config_filename, &config)) {
		exit(1);
	}
	if (test_config) {
		exit(0);
	}

start:
	fprintf(stderr, "\nCubemap " SERVER_VERSION " starting.\n");
	servers = new ServerPool(config.num_servers);

	CubemapStateProto loaded_state;
	struct timeval serialize_start;
	set<string> deserialized_stream_ids;
	map<string, Input *> deserialized_inputs;
	map<int, Acceptor *> deserialized_acceptors;
	if (state_fd != -1) {
		fprintf(stderr, "Deserializing state from previous process... ");
		string serialized;
		if (!read_tempfile(state_fd, &serialized)) {
			exit(1);
		}
		if (!loaded_state.ParseFromString(serialized)) {
			fprintf(stderr, "ERROR: Failed deserialization of state.\n");
			exit(1);
		}

		serialize_start.tv_sec = loaded_state.serialize_start_sec();
		serialize_start.tv_usec = loaded_state.serialize_start_usec();

		// Deserialize the streams.
		for (int i = 0; i < loaded_state.streams_size(); ++i) {
			servers->add_stream_from_serialized(loaded_state.streams(i));
			deserialized_stream_ids.insert(loaded_state.streams(i).stream_id());
		}

		// Deserialize the inputs. Note that we don't actually add them to any state yet.
		for (int i = 0; i < loaded_state.inputs_size(); ++i) {
			deserialized_inputs.insert(make_pair(
				loaded_state.inputs(i).stream_id(),
				create_input(loaded_state.inputs(i))));
		} 

		// Deserialize the acceptors.
		for (int i = 0; i < loaded_state.acceptors_size(); ++i) {
			deserialized_acceptors.insert(make_pair(
				loaded_state.acceptors(i).port(),
				new Acceptor(loaded_state.acceptors(i))));
		}

		fprintf(stderr, "done.\n");
	}

	// Find all streams in the configuration file, and create them.
	create_streams(config, deserialized_stream_ids, &deserialized_inputs);

	servers->run();

	vector<Acceptor *> acceptors = create_acceptors(config, &deserialized_acceptors);
	vector<Input *> inputs = create_inputs(config, &deserialized_inputs);
	
	// All deserialized inputs should now have been taken care of, one way or the other.
	assert(deserialized_inputs.empty());
	
	// Put back the existing clients. It doesn't matter which server we
	// allocate them to, so just do round-robin. However, we need to add
	// them after the mark pools have been set up.
	for (int i = 0; i < loaded_state.clients_size(); ++i) {
		servers->add_client_from_serialized(loaded_state.clients(i));
	}

	// Start writing statistics.
	StatsThread *stats_thread = NULL;
	if (!config.stats_file.empty()) {
		stats_thread = new StatsThread(config.stats_file, config.stats_interval);
		stats_thread->run();
	} else if (config.stats_interval != -1) {
		fprintf(stderr, "WARNING: 'stats_interval' given, but no 'stats_file'. No statistics will be written.\n");
	}

	signal(SIGHUP, hup);
	
	struct timeval server_start;
	gettimeofday(&server_start, NULL);
	if (state_fd != -1) {
		// Measure time from we started deserializing (below) to now, when basically everything
		// is up and running. This is, in other words, a conservative estimate of how long our
		// “glitch” period was, not counting of course reconnects if the configuration changed.
		double glitch_time = server_start.tv_sec - serialize_start.tv_sec +
			1e-6 * (server_start.tv_usec - serialize_start.tv_usec);
		fprintf(stderr, "Re-exec happened in approx. %.0f ms.\n", glitch_time * 1000.0);
	}

	while (!hupped) {
		usleep(100000);
	}

	// OK, we've been HUPed. Time to shut down everything, serialize, and re-exec.
	gettimeofday(&serialize_start, NULL);

	if (stats_thread != NULL) {
		stats_thread->stop();
	}
	for (size_t i = 0; i < acceptors.size(); ++i) {
		acceptors[i]->stop();
	}
	for (size_t i = 0; i < inputs.size(); ++i) {
		inputs[i]->stop();
	}
	servers->stop();

	fprintf(stderr, "Serializing state and re-execing...\n");
	CubemapStateProto state = collect_state(
                serialize_start, acceptors, inputs, servers);
	string serialized;
	state.SerializeToString(&serialized);
	state_fd = make_tempfile(serialized);
	if (state_fd == -1) {
		exit(1);
	}
	delete servers;

	if (!dry_run_config(argv[0], config_filename)) {
		fprintf(stderr, "ERROR: %s --test-config failed. Restarting old version instead of new.\n", argv[0]);
		hupped = false;
		goto start;
	}
	 
	char buf[16];
	sprintf(buf, "%d", state_fd);

	for ( ;; ) {
		execlp(argv[0], argv[0], config_filename.c_str(), "--state", buf, NULL);
		perror("execlp");
		fprintf(stderr, "PANIC: re-exec of %s failed. Waiting 0.2 seconds and trying again...\n", argv[0]);
		usleep(200000);
	}
}
